import wikipedia
import requests
from bs4 import BeautifulSoup
from duckduckgo_search import DDGS

from .utils.llm import llm_call
from .core.tool import Tool

# Symbolic Modules

def search_wikipedia(search_term:str):
    search_results = wikipedia.search(search_term)
    selected_search_result = search_results[0]
    page_summary = wikipedia.summary(selected_search_result)
    #page = wikipedia.page(search_term)
    return page_summary #page.content#, , page.links

search_wikipedia_tool_description = """Searches Wikipedia using the API to return the page summary, page content and list of links.
You must ensure that a page for your entered search term exists on Wikipedia.
Use the returned page content, pass it through the reasoning step with a question to get your answer."""

searchWikipediaTool = Tool(tool_name="Wikipedia Search",
                                 description=search_wikipedia_tool_description,
                                 action=search_wikipedia)




def reason(input_text:str):
    return llm_call([{"role" : "user", "content" : input_text}])

reason_tool_description = """Makes an LLM call to figure out an answer.
In the input text string, make sure that you include all the relevant context and clearly mention your question and the expected output from the LLM.
"""

reasonTool = Tool(tool_name = "Reason",
                        description = reason_tool_description,
                        action=reason)



def web_search_ddg(search_term:str):
    results = DDGS().text(search_term, max_results=5)
    return str(results)

web_search_ddg_tool_description = """Given a search query, searches the DuckDuckGo search engine.
Returns 5 search results with the title, URL and a few sentences of the body, of each search result.
The search results are only useful if, after searching, the agent uses the getWebpageContentTool to visit the most relevant URL, reads its content, and finds the answer to the user's question.
"""

webSearchTool = Tool(tool_name="Web Search",
                           description=web_search_ddg_tool_description,
                           action=web_search_ddg)



def get_webpage_content(url:str):
    try:
        response = requests.get(url)
        response.raise_for_status()
        soup = BeautifulSoup(response.content, 'html.parser')
        title = soup.title.string if soup.title else 'No title found'
        text_content = ' '.join(soup.stripped_strings)
        return f"{title}\n{text_content}"
    except requests.exceptions.RequestException as e:
        return "There was an error reading this webpage."

get_webpage_content_tool_description = """Given the URL to a webpage, returns the page content if it is accessible.
Before using this tool, you would usually have searched the web to have your search results' URLs ready.
The Web Search Tool's response only contains page titles and a few sentences. So, if these sentences do not answer the user's question, you must visit one of the webpage URLs to get the full text.
After using this tool, you can use the Reasoning tool to filter / summarise the text to generate the final answer for the user.
If a certain webpage's content is not accessible, the agent must try another search result URL.
"""

getWebpageContentTool = Tool(tool_name = "Get Webpage Content",
                        description = get_webpage_content_tool_description,
                        action=get_webpage_content)




#Human Feedback

def send_email(name:str, email_address:str, email_subject:str, email_body:str):
    pass #TODO

def poll_for_email_reply():
    pass #TODO


from ..utils.llm import llm_call

class Brain:
    def __init__(self, brain_system_prompt:str):
        self.brain_system_prompt = brain_system_prompt

    def reason(self, thought):
        messages = [{"role" : "system", "content" : self.brain_system_prompt},
                    {"role" : "user", "content" : thought}]
        return llm_call(messages)


class ToolSearchPrompt:
    def __init__(self, tools):
        self.tools = tools

    def generate_prompt(self):
        prompt = """
        You are part of an autonomous agent.
        You are provided with a list of tools the agent has. The user will provide a task they want to accomplish.
        You will return the name of the most suitable tool to be used for this task, as well as the values of the arguments to be plugged in.
        
        # Example tool:
        create_sheet: Creates a new worksheet.
        Arguments: sheet_name, company_name

        # Example user task:
        - Make a sheet named Sarthak for OpenAI.

        # Example response:
        {"tool_name" : "create_sheet", "sheet_name" : "Sarthak", "company_name" : "OpenAI"}

        Return your response strictly in the above JSON format. Only return this JSON and nothing else.
        Do not add any comments or explanations, or your response will not be parsed correctly.
        Below are the tools:
        - """

        prompt += "\n".join([f"- {tool.tool_name}: {tool.description}\nArguments: {', '.join(tool.args)}\n" for tool in self.tools])
        return prompt


from ..vectordb import get_embedding
class Tool:
    def __init__(self, tool_name, description, action, args=[]):
        self.tool_name = tool_name
        self.description = description
        #self.description_embedding = get_embedding(description)
        self.action = action
        self.args = args

class Instruction(Tool):
    def __init__(self, instruction_name, description, func, args, response_variable):
        self.instruction_name = instruction_name
        self.description = description
        self.func = func
        self.args = args
        self.response_variable = response_variable



from all_tools import send_email

class PointOfContact:
    def __init__(self, point_of_contact_name, communication_mode, communication_address) -> None:
        self.point_of_contact_name = point_of_contact_name
        self.communication_address = communication_address
        self.communication_mode = communication_mode

    def communicate(self, name, message_subject, message_body):
        if self.communication_mode == "email":
            return send_email(name=name,
                              email_address=self.communication_address,
                              email_subject=message_subject,
                              email_body=message_body)



from .tool_search import search_tool
from .core.tool import Tool, Instruction
from .core.planning import Planner
from .core.memory import ShortTermMemory, LongTermMemory
from .all_tools import reasonTool, webSearchTool, getWebpageContentTool
from .utils.llm import llm_call

class Agent:
    def __init__(self, brain, long_term_memory=None, planner=None, name="", purpose="", instructions=[], plan_steps=[], task="", verbose=2):
        self.name = name
        self.purpose = purpose
        self.brain = brain
        self.short_term_memory = ShortTermMemory()
        self.long_term_memory = long_term_memory
        self.tools = []
        self.planner = planner
        self.task = task
        self.plan_steps = plan_steps
        self.verbose = verbose
        self.add_tool(reasonTool)
        self.add_tool(webSearchTool)
        self.add_tool(getWebpageContentTool)

    def add_tool(self, tool):
        self.tools.append(tool)

    def add_point_of_contact(point_of_contact):
        #Add point of contact
        pass

    def execute_instruction(self, instructions_llm_response):
        instruction_function = [tool.action for tool in self.tools if tool.tool_name == instructions_llm_response['tool_name']][0]
        instruction_arguments = instructions_llm_response['arguments']
        result = instruction_function(**instruction_arguments)
        if self.verbose >= 2: print(f"""Performed {instruction_function.__name__} with args {str(instruction_arguments)}.\nResult: {result}""")
        return result




class PlannerAgent(Agent):
    def __init__(self, brain, long_term_memory=None, planner=None, name="", purpose="", instructions=[], plan_steps=[], task="", verbose=2):
        super().__init__(brain, long_term_memory, planner, name, purpose, instructions, plan_steps, task, verbose)

    def act(self):
        if self.plan_steps and not self.planner:
            self.planner = Planner(agent_goal=self.task)
            self.planner.from_preplanned_steps(self.plan_steps)

        elif not self.plan_steps and self.planner:
            self.planner.plan_task(self.tools, brain=self.brain)
            self.plan_steps = self.planner.agent_plan_steps

        elif not self.plan_steps and not self.planner:
            raise Exception("Provide either a Planner class object or plan_steps to the Agent class.")
        return self.execute_plan()
            
    def execute_plan(self):
        prev_step, result = "", ""
        for plan_step in self.plan_steps:
            plan_step = prev_step + plan_step
            instructions_llm_response = self.planner.convert_plan_step_to_instruction(plan_step, self.tools)
            result = self.execute_instruction(instructions_llm_response=instructions_llm_response)
            prev_step = f"""The previous step was:\n{plan_step}\nThe previous step's output was:\n{result}"""
        return result
    



class ReActAgent(Agent):
    """ https://arxiv.org/abs/2210.03629 """

    def __init__(self, brain, task, max_iterations, long_term_memory=None, name="", purpose="", instructions=[], verbose=2):
        super().__init__(brain=brain, task=task, long_term_memory=long_term_memory, name=name, purpose=purpose, instructions=instructions, verbose=verbose)
        self.max_iterations = max_iterations
        self.short_term_memory = ShortTermMemory()

        self.ReAct_preprompt = f"""You are part of an AI agent.
        The AI agent is built on this principle:
        {self.brain.brain_system_prompt}

        The agent is working on the below task/question given by the user:
        "{task}"

        The agent iterates in 3 steps: Think, Act, Observe.
        Remember that the agent can only iterate {str(self.max_iterations)} times -- after which, the current observation will be given to the user as the answer."""

        self.ReAct_postprompt = f"""If you think we already have the complete answer to the user's question or have finished their task, then present this answer or conclusion by responding in the following format:
        "BREAK | [Place the complete answer to user's question along with sources here]"
        """

    def create_ReAct_thought(self, iteration_count:int):
        tools = '\n\n'.join([f"- {tool.tool_name}\n{tool.description}" for tool in self.tools])
        
        create_ReAct_thought_system_prompt = f"""
        {self.ReAct_preprompt}
        You're current on iteration {str(iteration_count)}/{str(self.max_iterations)}.
        Your subtask is think about the best course of action for the agent.

        The agent has access to a set of tools it can use for the "action" in this iteration:
        {tools}
        
        {'This is the process the agent has completed so far:' + self.short_term_memory.formatted() if self.short_term_memory.is_not_empty() else ''}
        
        Think about the immediate next step the agent should take.
        {self.ReAct_postprompt}
        """
        return self.process_ReAct_llm_output(system_prompt=create_ReAct_thought_system_prompt)

    def create_ReAct_action(self, ReAct_thought):
        instructions_llm_response = Planner(agent_goal=ReAct_thought).convert_plan_step_to_instruction(ReAct_thought, self.tools)
        return instructions_llm_response

    def create_ReAct_observation(self, iteration_count, ReAct_thought, ReAct_action, instruction_execution_result):
        create_ReAct_observation_system_prompt = f"""
        {self.ReAct_preprompt}
        You're current on iteration {str(iteration_count)}/{str(self.max_iterations)}.
        
        {'This is the process the agent has completed so far:' + self.short_term_memory.formatted() if self.short_term_memory.is_not_empty() else ''}
        
        This was the agent's thought:
        {ReAct_thought}
        The agent decided to take this action:
        {ReAct_action}
        The action gave the following result:
        {instruction_execution_result}
        Your task is to understand the result of executing the response and write your observation.
        Remember that your response will be stored in the memory of the agent and will be used by subsequent agent iterations to finally answer the user's question. So, you must include any relevant information you find, in your response.
        {self.ReAct_postprompt}
        """
        return self.process_ReAct_llm_output(system_prompt=create_ReAct_observation_system_prompt)


    def summarise_ReAct_iteration(self, iteration_count, ReAct_thought, ReAct_action, ReAct_observation):
        return f""" Iteration {str(iteration_count)}
        This was the agent's thought:
        {ReAct_thought}
        The agent decided to take this action:
        {ReAct_action}
        After executing this action, the following result was observed:
        {ReAct_observation}
        """
    
    def process_ReAct_llm_output(self, system_prompt):
        messages = [{"role" : "system", "content" : system_prompt}]
        llm_response = llm_call(messages)
        if self.verbose >=2: print(llm_response)
        if llm_response.split("|")[0].strip() == "BREAK": return False, llm_response.split("|")[1].strip()
        else: return True, llm_response

    def compile_answer_from_short_term_memory(self):
        compile_answer_from_short_term_memory_system_prompt = f"""
        An AI agent has been working on the below task/question given by the user:
        "{self.task}"
        The AI agent is built on this principle:
        {self.brain.brain_system_prompt}

        The agent iterates in 3 steps: Think, Act, Observe.
        You are given the agent's work. You must use it as reference to answer the user's question.
        """
        messages = [{"role" : "system", "content" : compile_answer_from_short_term_memory_system_prompt},
                    {"role" : "user", "content" : self.short_term_memory.formatted()}]
        llm_response = llm_call(messages)
        return llm_response

    def act(self):
        for iteration_count in range(1, self.max_iterations+1):

            # Think
            continue_ReAct, ReAct_thought = self.create_ReAct_thought(iteration_count=iteration_count)
            if not continue_ReAct: return ReAct_thought

            # Act
            ReAct_action = self.create_ReAct_action(ReAct_thought=ReAct_thought)
            instruction_execution_result = self.execute_instruction(instructions_llm_response=ReAct_action)
            if self.long_term_memory: self.long_term_memory.add_new_memory(instruction_execution_result)

            # Observe
            continue_ReAct, ReAct_observation = self.create_ReAct_observation(iteration_count=iteration_count, ReAct_thought=ReAct_thought, ReAct_action=ReAct_action, instruction_execution_result=instruction_execution_result)
            if not continue_ReAct: return ReAct_observation

            self.short_term_memory.add_new_memory(self.summarise_ReAct_iteration(iteration_count=iteration_count, ReAct_thought=ReAct_action, ReAct_action=ReAct_action, ReAct_observation=ReAct_observation))
        
        #If the agent could not provide the answer in the given number of iterations, summarise the short term memory.
        return self.compile_answer_from_short_term_memory()


#all the relevant research findings get stored as constituent objects of this class.
from ..utils.llm import llm_call


class Memory:
    def __init__(self, memories=[]):
        self.memories = memories

    def add_new_memory(self, new_memory:str):
        self.memories += [new_memory]

    def is_not_empty(self):
        return True if self.memories else False
    
    def formatted(self):
        return "\n".join(self.memories)


class ShortTermMemory(Memory):
    def __init__(self, memories=[]):
        super().__init__(memories)
    

class LongTermMemory(Memory):
    def __init__(self, memories=[]):
        super().__init__(memories)


from .prompts import ToolSearchPrompt
from .utils.llm import llm_call
from .vectordb import vector_search, get_embedding

import json

def search_tool(tools, task, method):
    # Placeholder for different search mechanisms
    if method == "llm_call":
        tool_search_prompt = ToolSearchPrompt(tools).generate_prompt()
        messages = [{"role" : "system", "content" : tool_search_prompt},
                    {"role" : "user", "content" : task}]
        best_matching_tool_object = llm_call(messages, "gpt-3.5-turbo")
        best_matching_tool_object = json.loads(best_matching_tool_object)
        print(best_matching_tool_object)

        tool_name = best_matching_tool_object.pop("tool_name")
        best_matching_tool = [tool for tool in tools if tool.tool_name == tool_name][0]
        best_matching_tool.args = best_matching_tool_object

        return best_matching_tool

    elif method == "vector_search":
        task_embedding = get_embedding(task)
        best_matching_tool = vector_search(tools, task_embedding)
        return best_matching_tool
    



from setuptools import setup, find_packages

with open("README.MD", "r") as f:
    readme_content = f.read()

setup(
    name="nebulousai",
    version="0.1.1",
    packages=find_packages(),
    long_description=readme_content,
    long_description_content_type="text/markdown",
    install_requires=[
        "litellm==1.34.0",
        "wikipedia==1.4.0",
        "duckduckgo_search==6.1.0",

    ]
)


from litellm import embedding
from sklearn.metrics.pairwise import cosine_similarity

def get_embedding(text):
    return embedding(model="text-embedding-ada-002", input=text).data[0]['embedding']

def vector_search(tools, task_embedding):
    similarities = [cosine_similarity([task_embedding], [tool.description_embedding])[0][0] for tool in tools]
    best_tool = max(zip(tools, similarities), key=lambda x: x[1])[0]
    return best_tool


import re
import json
import inspect
from json import JSONDecodeError

from ..utils.llm import llm_call

class Planner:
    def __init__(self, agent_goal, refine_plan=False):
        self.agent_goal = agent_goal
        self.refine_plan = refine_plan
        self.agent_plan_steps = []

    #Decomposition into subgoals
    def decompose_tasks_into_subtasks(self, tools, brain):
                
        tools = '\n\n'.join([f"- {tool.tool_name}\n{tool.description}" for tool in tools])
        chain_of_thought_planning_prompt = f"""
        You are given a task the user wants to accomplish. These are the guidelines the user has set for you:
        {brain.brain_system_prompt}

        You have the following tools you can use to complete this task:
        {tools}

        Plan out a series of steps that allow you to complete this task. You can plan any number of tasks.
        In each step, name the tool to be used, and its input and expected output.
        Return your response in the below format. Do not add any comments in your response or it will not be parsed correctly!

        Step 1:
        Function, reason for using this function, input to function, expected output.

        Step 2:
        Function, reason for using this function, input to function, expected output.

        Step 3:
        Function, reason for using this function, input to function, expected output.
        """
        # https://arxiv.org/abs/2201.11903

        messages = [{"role" : "system", "content" : chain_of_thought_planning_prompt},
                    {"role" : "user", "content" : self.agent_goal}]
        
        llm_response_plan = llm_call(messages)
        steps = re.findall(r'Step \d+:(.*?)(?=Step \d+|$)', llm_response_plan, re.DOTALL)
        self.agent_plan_steps = [step.strip() for step in steps]


    def convert_plan_step_to_instruction(self, plan_step, tools, retries=5):
        tools_definition_str = ""
        for tool in tools:
            params = inspect.signature(tool.action).parameters.values()
            params_str = "\n".join([f'Name: {param.name}, Default: {param.default}, Kind: {param.kind}' for param in params])
            tools_definition_str += f"""\n\nTool: {tool.tool_name}\nDescription: {tool.description}\nArguments: {params_str}"""
        
        convert_plan_to_instructions_system_prompt = f"""
        You are part of an AI agent flow. The user will give you an action they want to perform.
        You also have a set of tools to perform it.
        FIgure out the appropriate tool and the input to be given to it.

        <available tools>
        - {tools_definition_str}
        </available tools>

        Return your response in the below JSON format. Do not add any comments in your response or it will not be parsed correctly!
        """ + """{"tool_name" : "toolName1",
         "arguments" : {"argument_1_name" : "argument_1_value", "argument_2_name" : "argument_2_value"}
        }"""
        
        convert_plan_to_instructions_system_prompt = convert_plan_to_instructions_system_prompt.replace("    ", "")
        try:
            messages = [{"role" : "system", "content" : convert_plan_to_instructions_system_prompt},
                        {"role" : "user", "content" : plan_step}]
            
            instructions_llm_response = llm_call(messages)
            instructions_llm_response_parsed = json.loads(instructions_llm_response)
        except JSONDecodeError:
            if retries > 0: return self.convert_plan_step_to_instruction(plan_step, tools, retries-1)
        return instructions_llm_response_parsed
    

    #Reflection and refinement
    def reflect_upon_and_refine_plan_llm_call(plan, tools):
        reflect_upon_and_refine_plan_system_prompt = ""
        messages = [{"role" : "system", "content" : reflect_upon_and_refine_plan_system_prompt},
                    {"role" : "user", "content" : plan}]
        
        instructions_llm_response = llm_call(messages)
        instructions_llm_response = instructions_llm_response.split("\n")
        return instructions_llm_response
    

    def plan_task(self, tools, brain):
        self.decompose_tasks_into_subtasks(tools=tools, brain=brain)
        if self.refine_plan: self.reflect_upon_and_refine_plan_llm_call()

    
    def from_preplanned_steps(self, plan_steps):
        self.agent_plan_steps = plan_steps




class TreeOfThoughtPlanner:
    def __init__(self, agent_goal, refine_plan=False):
        self.agent_goal = agent_goal
        self.refine_plan = refine_plan
        self.agent_plan_steps = []

    def decompose_tasks_into_subtasks(self, tools):
        tree_of_thought_prompt = """"""
        # exploring multiple reasoning possibilities at each step. It first decomposes the problem into multiple thought steps and generates multiple thoughts per step, creating a tree structure. The search process can be BFS (breadth-first search) or DFS (depth-first search) with each state evaluated by a classifier (via a prompt) or majority vote.
        # https://arxiv.org/abs/2305.10601
        pass